/**************************************************************************************************

AtMost.C -- (C) Daniel Andrn, 2005

Defines the constraint AtMost supported by UmSat.

**************************************************************************************************/

#include "AtMost.h"
#include "Solver.h"
#include "Sort.h"

//=================================================================================================
// AtMost constraint


// Pre-condition: All variables must be distinct and unbound. (To lift this pre-condition,
// any fact immediately derivable from the new constraint should be enqueued by 'enqueue()'.
// If the constraint is already satisfied under the current top-level assignment, it should
// not be constructed at all.)
//
bool AtMost_new(Solver& S, const vec<Lit>& ps, int n, AtMost*& out_atmost)
{
    assert(S.decisionLevel() == 0);
    if (&out_atmost != NULL) out_atmost = NULL;

    if(n < 0) return false;

    if(n == 0) {
      for(int i = 0; i < ps.size(); ++i) 
	if(!S.enqueue(~ps[i])) return false;
      return true;
    }
    
    if(ps.size() == 0) return true;
    
    void* mem    = (void*)xmalloc<char>(sizeof(AtMost) + sizeof(Lit) * ps.size());
    AtMost* out  = new (mem) AtMost;
    out->size    = ps.size();
    out->n       = n;
    out->counter = 0;
    for (int i = 0; i < ps.size(); i++) out->lits[i] = ps[i];
    for (int i = 0; i < ps.size(); i++) S.watches[index(ps[i])].push(out);

    if (&out_atmost != NULL) out_atmost = out;
    return true;
}


void AtMost::remove(Solver& S, bool just_dealloc) {
    if (!just_dealloc)
        for (int i = 0; i < size; i++)
            removeWatch(S.watches[index(lits[i])], this);
    xfree(this);
}


bool AtMost::simplify(Solver& S) {
    return false; }

bool AtMost::propagate(Solver& S, Lit p, bool& keep_watch)
{
  // S.watches[index(p)].push(this);
  keep_watch = true;

  // To collect some statistics!
  ++S.stats.prop[0];
  ++S.stats.propm;

  // Can this happen?
  if (counter == n) return false;
    
  ++counter;
  S.undos[var(p)].push(this);
    
  // If no more can be true, enqueue the negation of the rest:
  if (counter == n) {
    int cnt = 0;
    for (int i = 0; i < size; ++i) {
      if (S.value(lits[i]) == l_Undef && !S.enqueue(~lits[i], this)) return false;
      if (S.value(lits[i]) == l_True) ++cnt;
    }
    if(cnt > n) return false; // If more than n lits l_True => conflict
  }
    
  return true;
}


void AtMost::undo(Solver& S, Lit p) {
  // S.watches[index(p)].push(this); // Hack
    counter--; }


void AtMost::calcReason(Solver& S, Lit p, vec<Lit>& out_reason)
{
    int     c = (p == lit_Undef) ? -1 : 0;
    for (int i = 0; i < size; i++){
        if (S.value(lits[i]) == l_True){
            out_reason.push(lits[i]);
            if (++c == n) return;
        }
    }
}
