#ifndef ConstraintsDB_h
#define ConstraintsDB_h

#include "SolverTypes.h"
#include "Clause.h"
#include "Queue.h"
#include "VarOrder.h"

//=================================================================================================

struct SearchParams {
    double  var_decay, clause_decay, random_var_freq;    // (reasonable values are: 0.95, 0.999, 0.02)    
    SearchParams(double v = 1, double c = 1, double r = 0) : var_decay(v), clause_decay(c), random_var_freq(r) { }
};

//=================================================================================================

template <class T>
class ConstraintsDB {
  vec<T> constrs;       // List of problem constraints.
 public:
  int size() const { return constrs.size(); };
  void remove(Solver& S, int i, bool flag = false) { constrs[i]->remove(S, flag); };
  void add(T c) { constrs.push(c); };
  void simplifyDB(Solver& S);
};

//=================================================================================================

template<>
class ConstraintsDB<LClause *> {
  double              cla_inc;        // Amount to bump next clause with.
  double              cla_decay;      // INVERSE decay factor for clause activity: stores 1/decay.
  
  vec<double>         activity;       // A heuristic measurement of the activity of a variable.
  double              var_inc;        // Amount to bump next variable with.
  double              var_decay;      // INVERSE decay factor for variable activity: stores 1/decay. Use negative value for static variable order.
  VarOrder            order;          // Keeps track of the decision variable order.
  
  vec<LClause*> constrs;       // List of problem constraints.
 public:
  void add(LClause *c) { constrs.push(c); };
  void remove(Solver& S, int i, bool flag = false) { constrs[i]->remove(S, flag); };

  int size() const { return constrs.size(); };

  void newVar(bool decision_var) {
    activity.push(0);
    order.newVar(decision_var);
  }

  inline void undo(Var x) {
    order.undo(x);
  }

  inline Var select(Solver& S, double random_var_freq) {
    return order.select(S, random_var_freq);
  }

  vec<Var> getOrder(Solver& S) { return order.getOrder(S); }

  void simplifyDB(Solver& S);
  void reduceDB(Solver& S);

  void setParams(const SearchParams& params) {
    var_decay = 1 / params.var_decay;
    cla_decay = 1 / params.clause_decay;
  }

  // Activity:
  //
  void    varBumpActivity(Lit p) {
    if (var_decay < 0) return;     // (negative decay means static variable order -- don't bump)
    if ( (activity[var(p)] += var_inc) > 1e100 ) varRescaleActivity();
    order.update(var(p)); }
  void    varDecayActivity(void) { if (var_decay >= 0) var_inc *= var_decay; }
  void    varRescaleActivity(void);
  
  void    claBumpActivity(LClause* c) { if ( (c->activity() += cla_inc) > 1e20 ) claRescaleActivity(); }
  void    claDecayActivity(void) { cla_inc *= cla_decay; }
  void    claRescaleActivity(void);
  
  ConstraintsDB(vec<char>& assigns) : 
    cla_inc     (1)
    , cla_decay (1)
    , var_inc   (1)
    , var_decay (1)
    , order     (assigns, activity)
    {};
};

//=================================================================================================

/*_________________________________________________________________________________________________
|                                                                                                  
|  simplifyDB : [void]  ->  [void]                                                                 
|                                                                                                  
|  Description:                                                                                    
|    Simplify all constraints according to the current top-level assigment (redundant constraints  
|    may be removed altogether).                                                                   
|________________________________________________________________________________________________@*/
template <class T>
void ConstraintsDB<T>::simplifyDB(Solver& S)
{
  int j = 0;
  for (int i = 0; i < size(); i++){
    if (constrs[i]->simplify(S))
      remove(S,i);
    else
      constrs[j++] = constrs[i];
  }
  constrs.shrink(constrs.size()-j);
}

#endif
