/**************************************************************************************************

SClause.C -- (C) Niklas Een, Niklas Srensson, 2004

Defines the constraints supported by MiniSat (clauses being the core constraint). 

**************************************************************************************************/

#include "Clause.h"
#include "Solver.h"
#include "Sort.h"

//=================================================================================================
// SClause constraint:

// Returns FALSE if top-level conflict detected (must be handled); TRUE otherwise.
// 'out_clause' may be set to NULL if clause is already satisfied by the top-level assignment.
//
bool SClause_new(Solver& S, const vec<Lit>& ps_, SClause*& out_clause)
{
    vec<Lit>    qs;
    if (&out_clause != NULL) out_clause = NULL;

    assert(S.decisionLevel() == 0);
    ps_.copyTo(qs);             // Make a copy of the input vector.
    
    // Remove false literals:
    for (int i = 0; i < qs.size();){
      if (S.value(qs[i]) != l_Undef){
	if (S.value(qs[i]) == l_True)
	  return true;    // SClause always true -- don't add anything.
	else
	  qs[i] = qs.last(),
	    qs.pop();
      }else
	i++;
    }
    
    // Remove duplicates:
    sortUnique(qs);
    for (int i = 0; i < qs.size()-1; i++){
      if (qs[i] == ~qs[i+1])
	return true;        // SClause always true -- don't add anything.
    }

    const vec<Lit>& ps = qs;     // 'ps' is now the (possibly) reduced vector of literals.

    if (ps.size() == 0)
        return false;
    else if (ps.size() == 1)
        return S.enqueue(ps[0]);
    else {
        // Allocate clause:
        assert(sizeof(Lit)   == sizeof(unsigned));
	assert(sizeof(float) == sizeof(unsigned));
	void*   mem = xmalloc<char>(sizeof(SClause) + sizeof(unsigned)*(ps.size()));
	SClause* c   = new (mem) SClause;

        c->setSize(ps.size());
        for (int i = 0; i < ps.size(); i++) c->data[i] = ps[i];

	c->setIdx(2);

        // Store clause:
        S.watches[index(~c->data[0])].push(c);
        S.watches[index(~c->data[1])].push(c);
        if (&out_clause != NULL) out_clause = c;

        return true;
    }
}

void SClause::remove(Solver& S, bool just_dealloc)
{
    if (!just_dealloc){
        removeWatch(S.watches[index(~data[0])], this);
        removeWatch(S.watches[index(~data[1])], this); }
    xfree(this);
}


// Can assume everything has been propagated! (esp. the first two literals are != l_False, unless
// the clause is binary and satisfied, in which case the first literal is true)
// Returns True if clause is satisfied (will be removed), False otherwise.
//
bool SClause::simplify(Solver& S)
{
    assert(S.decisionLevel() == 0);
    for (int i = 0; i < size(); i++){
        if (S.value(data[i]) == l_True)
            return true;
    }
    return false;
}


// 'p' is the literal that became TRUE
bool SClause::propagate(Solver& S, Lit p, bool& keep_watch)
{
    // To collect some statistics!
    ++S.stats.prop[0];
    if( size() <= NPROP ) ++S.stats.prop[size()-1];

    // Make sure the false literal is data[1]:
    Lit     false_lit = ~p;
    if (data[0] == false_lit)
        data[0] = data[1], data[1] = false_lit;
    assert(data[1] == false_lit);

    // If 0th watch is true, then clause is already satisfied.
    if (S.value(data[0]) == l_True){
        keep_watch = true;
        return true; }

    // Look for new watch:

    /***
    for (int i = 2; i < size(); i++){
        if (S.value(data[i]) != l_False){
            data[1] = data[i], data[i] = false_lit;
            S.watches[index(~data[1])].push(this);
            return true; } }
    ***/

    setIdx(2);
    for (int i = getIdx(); i < size(); i++){
        if (S.value(data[i]) != l_False){
            data[1] = data[i], data[i] = false_lit;
            S.watches[index(~data[1])].push(this);
	    setIdx(i+1);
            return true; } }

    for (int i = 2; i < getIdx(); i++){
        if (S.value(data[i]) != l_False){
            data[1] = data[i], data[i] = false_lit;
            S.watches[index(~data[1])].push(this);
	    setIdx(i+1);
            return true; } }

    // SClause is unit under assignment:
    keep_watch = true;

    return S.enqueue(data[0], this);
}


// Can assume 'out_reason' to be empty.
// Calculate reason for 'p'. If 'p == lit_Undef', calculate reason for conflict.
//
void SClause::calcReason(Solver& S, Lit p, vec<Lit>& out_reason)
{
    assert(p == lit_Undef || p == data[0]);
    for (int i = ((p == lit_Undef) ? 0 : 1); i < size(); i++)
        assert(S.value(data[i]) == l_False),
        out_reason.push(~data[i]);
}
