/**************************************************************************************************

jmf.C -- (C) Daniel Andrn, 2004

Defines the constraints supported by MiniSat.

**************************************************************************************************/

#include "jmf.h"
#include "Solver.h"
#include "debug.h"

#define HUGE 10000

//=================================================================================================
// jmf constraint:


// Returns FALSE if top-level conflict detected (must be handled); TRUE otherwise.
// 'out_clause' may be set to NULL if clause is already satisfied by the top-level assignment.
//
bool jmf_new(Solver& S, const vec<Lit>& ps, const vec<Lit>& qs, jmf*& out_clause)
{
    if (&out_clause != NULL) out_clause = NULL;

    assert(S.decisionLevel() == 0);

    if (ps.size() == 0)
        return true;
    //else if (ps.size() == 1)
    //  return S.enqueue(~ps[0]);
    else{
      assert(ps.size() == qs.size());
      // Allocate clause:
      assert(sizeof(Lit)   == sizeof(unsigned));
      assert(sizeof(float) == sizeof(unsigned));
      void* mem = xmalloc<char>(sizeof(jmf) + sizeof(unsigned)*(ps.size() + qs.size()));
      jmf*  x   = new (mem) jmf;

      x->size = ps.size();
      for (int i = 0; i < ps.size(); ++i) x->lits[i] = ps[i];
      for (int i = 0; i < qs.size(); ++i) x->lits[i+x->size] = qs[i];

      S.watches[index(~x->lits[0])].push(x);
      S.watches[index(x->lits[x->size])].push(x);
      x->mw = 1;

      if (&out_clause != NULL) out_clause = x;

      return true;
    }
}


void jmf::remove(Solver& S, bool just_dealloc)
{
  if (!just_dealloc){
    for(int i = 0; i < 2*size; ++i) {
      removeIfWatch(S.watches[index(lits[i])], this);
      removeIfWatch(S.watches[index(~lits[i])], this);
    }
  }
  xfree(this);
}


bool jmf::simplify(Solver& S)
{
  assert(S.decisionLevel() == 0);
  return false;
}


inline void jmf::add_watch(Solver& S, int i) {
  for(int j = mw; j < i; ++j) {
    S.watches[index(~lits[j])].push(this);
    S.watches[index(lits[j+size])].push(this);
  }
  mw = i;
}

// 'p' is the literal that became TRUE
inline bool jmf::propagate(Solver& S, Lit p, bool& keep_watch)
{
  // To collect some statistics!
  ++S.stats.prop[0];
  ++S.stats.propj;

  keep_watch = true;
  int i;
  for(i = 0; i < size; ++i) {
    if(S.value(lits[i]) == l_False) {
      if(S.value(lits[i+size]) == l_True) {
	if(i > mw) add_watch(S, i);
	return false;
      }
      else S.enqueue(~lits[i+size],this);
    } else {
      if(S.value(lits[i+size]) != l_True) break;
      else S.enqueue(lits[i],this);
    }
  }
    
  if(i == size || (S.value(lits[i]) == l_True && S.value(lits[i+size]) == l_False)) {
    if(i > mw) add_watch(S, i);
    return true;
  }
  
  if(i > mw) add_watch(S, i);
  return true;
}

void jmf::undo(Solver& S, Lit p) {}

// Can assume 'out_reason' to be empty.
// Calculate reason for 'p'. If 'p == lit_Undef', calculate reason for conflict.
//
void jmf::calcReason(Solver& S, Lit p, vec<Lit>& out_reason)
{
  //debugprint(10,0,AFTER,"Entering jmf::calcReason ("L_LIT" = %s)",L_lit(p),L_val(p,S));

  int i;
  for(i = 0; i < size; ++i) {
    if(S.value(lits[i]) == l_False) {
      if(S.value(lits[i+size]) == l_True) {
	out_reason.push(~lits[i]);
	out_reason.push(lits[i+size]);
	return;
      } else {
	out_reason.push(~lits[i]);
	if(~lits[i+size] == p) return;
      }
    } else {
      if(S.value(lits[i+size]) != l_True) {
	if(S.value(lits[i]) == l_True && S.value(lits[i+size]) == l_False) return;
	break;
      } else {
	out_reason.push(lits[i+size]);
	if(lits[i] == p) return;
      }
    }
  }

  return;

  /***
  debugprint(10,0,NONE,"reason =");
  for(int i = 0; i < out_reason.size(); ++i) debugprint(10,0,NONE," "L_LIT,L_lit(out_reason[i]));
  debugprint(10,0,AFTER,"");  
  ***/
}
