/**************************************************************************************************

Main.C -- (C) Daniel Andrn, 2005

Reads indatafiles file and apply the SAT-solver to it.

**************************************************************************************************/

#ifdef HAVE_CONFIG_H
#    include <config.h>
#endif

#include <ctime>
#include <signal.h>
#include <unistd.h>
#include <iostream>
#include <vector>
#include <string>
#include <ostream>
#include <iterator>
#include "Solver.h"
#include "Parser.h"
#include "Stats.h"
//#include <boost/program_options.hpp>
//#include "test.h"

//namespace po=boost::program_options;
using namespace std;

//=================================================================================================

const string version(PACKAGE_VERSION);
const string progname(PACKAGE_NAME);

// A helper function to simplify the main part.
template<class T>
ostream& operator<<(ostream& os, const vector<T>& v)
{
  copy(v.begin(), v.end(), ostream_iterator<T>(cout, " "));
  return os;
}

//=================================================================================================

// class returnvalue;

class returnvalue {
public:
  static const returnvalue* UNKNOWN;
  static const returnvalue* SATISFIABLE;
  static const returnvalue* UNSATISFIABLE;
  const string& name (void) const { return m_name;  };
  const int     value(void) const { return m_value; };
private:
  returnvalue(string name, int value):
    m_name(name), m_value(value)
  {};
  string const m_name;
  int const    m_value;
};

const returnvalue* returnvalue::UNKNOWN       = new returnvalue("s UNKNOWN\n", 0);
const returnvalue* returnvalue::SATISFIABLE   = new returnvalue("s SATISFIABLE\n", 10);
const returnvalue* returnvalue::UNSATISFIABLE = new returnvalue("s UNSATISFIABLE\n", 20);

//=================================================================================================

class options {
public:
  int            verbosity;
  int            time;
  vector<string> infiles;
  bool           all;
  bool           count;
  bool           model;
  int            model_size;

  options::options():
    verbosity(0), time(-1), all(false), count(false), model(false), model_size(-1)
  {};
  
  int get_options(int argc, char *argv[]);
private:
};

int options::get_options(int argc, char *argv[]) {

  if(argc < 8) {
    printf("\n\
Usage: %s time verbosity all count model model_size files...\n\
\n\
time        Time to run in seconds. Less than 0 for infinity.\n\
verbosity   Verbosity level.\n\
all         Find all solutions? (Boolean t/f)\n\
count       Count solutions?    (Boolean t/f)\n\
model       Write model?        (Boolean t/f)\n\
model_size  Number of variables in the model.\n\
files...    Problem files.\n\n\
", argv[0]);
    return 0;
  }
  
  time = atoi(argv[1]);

  verbosity = atoi(argv[2]);

  all = argv[3][0] == 't' ? true : false;

  count = argv[4][0] == 't' ? true : false;
  if(count) all = true;

  model = argv[5][0] == 't' ? true : false;

  model_size = atoi(argv[6]);

  for(int i=7; i < argc; i++)
    infiles.push_back(argv[i]);

  return 1;
}


/*
int options::get_options(int argc, char *argv[]) {
  try{
    po::options_description desc("Optoins");
    desc.add_options()
      ("help,h",                                                       "produces a help message")
      ("time,t",    po::value<int>(&time)->      default_value(-1),    "sets maximal runningtime (in seconde), "
       "set to negative value to run forever")
      ("all,a",     po::bool_switch(&all)->      default_value(false), "finds all solutions")
      ("count,c",   po::bool_switch(&count)->    default_value(false), "count the solutions (implies --all)")
      ("model,q",   po::bool_switch(&model)->    default_value(false), "supresses printing of model(s)")
      ("verbose,v", po::value<int>(&verbosity)-> default_value(0),     "sets verbosity")
      ("model,m",   po::value<int>(&model_size),                       "number of variables in the model")
      ("version",                                                      "prints version")
      ;

    po::options_description inf("Infile");
    inf.add_options()
      ("input-files",po::value< vector<string> >(&infiles), "infiles")
      ;
    
    po::options_description options("Optons");
    options.add(desc).add(inf);

    po::positional_options_description p;
    p.add("input-files",-1);
    
    po::variables_map vm;
    po::store(po::command_line_parser(argc,argv).
	      options(options).
	      positional(p).run(),
	      vm);
    po::notify(vm);
    
    if (vm.count("help")) {
      cout << "Usage: UmSat [options] [infile(s)...]\n";
      cout << desc << "\n";
      return 1;
    }
    
    if (vm.count("version")) {
      cout << "c " << progname << " version " << version << "\n";
      return 0;
    }

    if(count) all = true; // count => all

  } catch(std::exception& e) {
	  std::cout << e.what() << "\n";
	  return 1;
  } catch (...) {
    std::cout << "Exception of unknown type!\n";
    return 2;
  }
};
*/

//=================================================================================================

void write_model(Solver &S, int model_size) {
  if(model_size > 0) {
    cout << "v " << model_size << " ";
    for( int i = 0; i < model_size-1; ++i) {
      if(S.model[i] == l_Undef) continue;
      int j = S.model[i]==l_True ? (i+1) : -(i+1);
      cout << j << " ";
    }
    int i = model_size-1;
    if(S.model[i] != l_Undef) {
      int j = S.model[i]==l_True ? (i+1) : -(i+1);
      cout << j;
    }
    cout << "\n";
  }
  return;
}

//=================================================================================================

void write_order(Solver &S) {
  vec<int> list = S.getOrder();
  cout << "c Order ";
  for( int i = 0; i < list.size()-1; ++i) {
    cout << list[i]+1 << " ";
  }
  int i = list.size()-1;
  cout << list[i]+1 << "\n";
  return;
}

//=================================================================================================

void add_solution(Solver &S, int model_size) {
  vec<Lit> nomodel;
  for(int i = 0; i < model_size; ++i) nomodel.push(S.model[i]==l_True ? ~Lit(i) : Lit(i));
  S.addSClause(nomodel);
  S.simplifyDB();
  return;
}

//=================================================================================================

int main(int argc, char** argv)
{
  options op;
  if(!op.get_options(argc,argv)) return 0;

  Solver      S;
  parser      p(S);
  
  /*
  for(int i=1; i < argc; i++)
    op.infiles.push_back(argv[i]);
  */

  p.parse(op.infiles);
 
  // op.all = true; // HACK!!!
  // op.model_size = 84; // HACK!!!
 
  if (!S.okay()) {
    cout << returnvalue::UNSATISFIABLE->name();
    return returnvalue::UNSATISFIABLE->value();
  }

  if(op.time > 0) alarm(op.time); // (s)

  S.verbosity = op.verbosity;

  if(op.model_size < 0) op.model_size = S.nVars();
  if(op.model_size > S.nVars()) op.model_size = S.nVars();

  int returnvalue;
  int cnt = 0;
  lbool ans;
  do{
    ans = S.solve();
    
    if(ans == l_Undef) {
      cout << returnvalue::UNKNOWN->name();
      if(op.model) write_order(S);
      if(op.model) write_model(S,S.nVars());
      returnvalue = returnvalue::UNKNOWN->value();
    } else if (ans == l_True) {
      ++cnt;

      cout << returnvalue::SATISFIABLE->name();
      returnvalue = returnvalue::SATISFIABLE->value();
       
      if(op.all) add_solution(S,op.model_size);
      if(op.model) write_model(S,op.model_size);
      if(!S.okay()) {
	cout << returnvalue::UNSATISFIABLE->name();
	returnvalue = returnvalue::UNSATISFIABLE->value();
	break;
      }
    } else { // ans == l_False
      cout << returnvalue::UNSATISFIABLE->name();
      returnvalue = returnvalue::UNSATISFIABLE->value();
    }
    
  } while(op.all && ans == l_True);
  
  if(S.verbosity > 1 ) printStats(S.stats, cpuTime());
  if(op.count) cout << "c " << cnt << "\n";
  return returnvalue;
}
