/**************************************************************************************************

Parser.C -- (C) Daniel Andrn, 2004, 2005

Defines a parser for UmSat.

**************************************************************************************************/

#include <string.h>
#include <vector>
#include "SolverTypes.h"
#include "Parser.h"
#include "readers.h"

using namespace std;

//=================================================================================================
// Parser

void parser::file (void) {
  skipOWS();
  // ('c'|'C') [^\n]* '\n'
  while(isC()) { skipC(); skipNEOL(); skipMEOL(); skipOWS(); };
  skipOEWS();

  pline();
  skipOEWS();

  // <vector>*
  while(isM09()) { handler->read(sol, *this); skipOEWS(); }
  skipOEWS();

  delete handler;
}

const char *parser::ptypeArray[] = {"UNKNOWN", "CNF", "AtMost", "Xor", "Imply", "jmf", "Lex"};

void parser::pline (void) {
  // 'p' <type> <pint> <int0> '\n'

  skipOWS();
  skipP();
  skipMWS();
  ftype = type();
  if(ftype == unknown) { Error("Unknown filetype.\n"); return; }
  skipMWS();
  nvar = pint();
  skipMWS();
  ncon = int0();
  skipOWS();

  skipMEOL();
}

parser::ptype parser::type (void) {
  char line[21];
  int c = 0;
  while( !isWS() && c < 20) { line[c] = lh | 0x20; ++c; skip(); }
  line[c] = 0;
  while( !isWS() ) skip();

  if(strncmp(line,"cnf",3)==0) {
    this->handler = new readcnf;
    return cnf;
  }
  if(strncmp(line,"xor",3)==0) {
    handler = new readXor;
    return Xor;
  }
  if(strncmp(line,"atm",3)==0) {
    handler = new readAtMost;
    return AtMost;
  }
  if(strncmp(line,"imp",3)==0) {
    handler = new readImply;
    return Imply;
  }
  if(strncmp(line,"jmf",3)==0) {
    handler = new readjmf;
    return jmf;
  }
  if(strncmp(line,"lex",3)==0) {
    handler = new readLex;
    return Lex;
  }

  return unknown;
}

int parser::pint (void) {
  // <pint> = [1-9][0-9]*
  int i;

  i = lh - '0';
  skip19();

  while(is09()) { 
    i = i * 10 + lh - '0'; 
    skip09(); 
  }
  
  return i;
}

int parser::int0 (void) {
  // <int> = [0-9]+
  int i;

  i = lh - '0';
  skip09();

  while(is09()) { 
    i = i * 10 + lh - '0'; 
    skip09(); 
  }
  
  return i;
}

int parser::sint (void) {
  // <sint> = -?<pint>
  int sign(1);
  if(isM()) { sign = -1; skipM(); }
  return sign * pint();
}

void parser::vector (vec<Lit>& v) {
  // <vector> = <lit>* '0'
  while(isM19()) {
    v.push(lit());
    skipMEWS();
  }
  skip0();
}

Lit parser::lit (void) {
  // <lit> = <sint>
  int i = sint();
  while(abs(i) > sol.nVars()) sol.newVar();
  return (i > 0 ? Lit(abs(i)-1) : ~Lit(abs(i)-1));
}

void parser::parsef (const string& name) {
  fname = name.c_str();
  inf = fopen(fname,"r");
  if(inf == 0) {
    Error("Can not open file '%s'\n", fname);
    return;
  }
  lh = getc(inf);
  line = 1;
  file();
  fclose(inf);
  inf = 0;
}

void parser::parse (const std::vector<std::string>& infiles) {
  for(std::vector<std::string>::const_iterator i = infiles.begin(); i != infiles.end(); ++i) {
    parsef(*i);
  }
}
