/**************************************************************************************************

parser.h -- (C) Daniel Andrn, 2004

Defines a parser for MiniSat.

**************************************************************************************************/

#ifndef __parser_h__
#define __parser_h__

#include <stdio.h>
#include <stdarg.h>
#include <vector>
#include <string>
#include "Global.h"
#include "Solver.h"
#include "debug.h"

class reader;


class parser {
 private:
  enum ptype { unknown, cnf, AtMost, Xor, Imply, jmf, Lex };
  static const char *ptypeArray[7];
  
  FILE *inf;
  const char *fname;
  int line, col;
  char lh;
  int nvar;
  int ncon;
  Solver& sol;
  ptype ftype;
  reader* handler;

  inline bool isC    (void) { return (lh == 'c'  || lh == 'C' ); }
  inline bool isP    (void) { return (lh == 'p'  || lh == 'P' ); }
  inline bool is09   (void) { return (lh >= '0'  && lh <= '9' ); }
  inline bool is19   (void) { return (lh >= '1'  && lh <= '9' ); }
  inline bool is0    (void) { return (lh == '0'  || lh == EOF ); }
  inline bool isM    (void) { return (lh == '-'               ); }
  inline bool isM19  (void) { return (isM()      || is19()    ); }
  inline bool isM09  (void) { return (isM()      || is09()    ); }
  inline bool isWS   (void) { return (lh == '\t' || lh == ' ' ); }
  inline bool isEOL  (void) { return (lh == '\n' || lh == '\r'); }
  inline bool isEWS  (void) { return (isWS()     || isEOL()   ); }
  inline bool isNEOL (void) { return (!isEOL()                ); }

  inline void skipC    (void) { if( isC()   ) skip(); else Error("Expexted 'C' found '%c'\n", lh); }
  inline void skipP    (void) { if( isP()   ) skip(); else Error("Expexted 'P' found '%c'\n", lh); }
  inline void skip09   (void) { if( is09()  ) skip(); else Error("Expexted 0-9 found '%c'\n", lh); }
  inline void skip19   (void) { if( is19()  ) skip(); else Error("Expexted 1-9 found '%c'\n", lh); }
  inline void skip0    (void) { if( is0()   ) skip(); else Error("Expexted 0 found '%c'\n",   lh); }
  inline void skipM    (void) { if( isM()   ) skip(); else Error("Expexted '-' found '%c'\n", lh); }
  inline void skipWS   (void) { if( isWS()  ) skip(); else Error("Expexted white space found '%c'\n", lh); }
  inline void skipEWS  (void) { if( isEWS() ) skip(); else Error("Expexted white space or end of line found '%c'\n", lh); }
  inline void skipEOL  (void) { if( isEOL() ) skip(); else Error("Expexted end of line found '%c'\n", lh); }

  inline void skipOWS  (void) { while( isWS()   ) skip(); }
  inline void skipOEWS (void) { while( isEWS()  ) skip(); }
  inline void skipOEOL (void) { while( isEOL()  ) skip(); }
  inline void skipNEOL (void) { while( isNEOL() ) skip(); }
  inline void skipMWS  (void) { skipWS();  skipOWS();  }
  inline void skipMEWS (void) { skipEWS(); skipOEWS(); }
  inline void skipMEOL (void) { skipEOL(); skipOEOL(); }

  void file ();
  void pline (void);
  ptype type (void);
  int pint (void);
  int int0 (void);
  void vector (vec<Lit>& v);
  Lit parser::lit (void);
  int sint (void);
  void parsef(const std::string&);

  inline void Error(const char* fmt, ...) {
    va_list args;
    va_start(args, fmt);
    fprintf (stdout, "\nIn file %s at line %d column %d : ", fname, line, col);
    vfprintf(stdout, fmt, args);
    va_end(args);
  }
  
  void skip (void) {
    isEOL() ? line++, col = 0 : col++;
    int c = getc(inf);
    if(c == EOF) 
      lh = 0;
    else 
      lh = c; 
  }

  friend class readXor;
  friend class readAtMost;
  friend class readImply;
  friend class readjmf;
  friend class readLex;
  friend class readcnf;


 public:
  parser(Solver& S) : inf(0), line(1), col(0), lh(0), sol(S) {}
  ~parser () { if(inf != 0) fclose(inf); }
  void parse (const std::vector<std::string>&);
};

#endif
