/**************************************************************************************************

Solver.C -- (C) Niklas Een, Niklas Srensson, 2004

A simple Chaff-like SAT-solver with support for incremental SAT.

**************************************************************************************************/

#include "Solver.h"
#include "Sort.h"
#include "ConstraintsDB.h"
#include <cmath>
#include <signal.h>

//=================================================================================================
// Timing control:

static sig_atomic_t graceful_quit = 0; // Flag to end it all

static void sig_alrm(int signo) {
  graceful_quit = 1;
  return;
}

//=================================================================================================
// Debug:


// For derivation output (verbosity level 5)
#define L_IND    "c %-*d"
#undef L_ind
#define L_ind    decisionLevel()*3+3,decisionLevel()
#define L_LIT    "%sx%d"
#define L_lit(p) sign(p)?"~":"", var(p)

// Just like 'assert()' but expression will be evaluated in the release version as well.
inline void check(bool expr) { assert(expr); }

//=================================================================================================
// Constructor:

Solver::Solver(void) : 
  ok                  (true)
  //, cla_inc           (1)
  //, cla_decay         (1)
  //, var_inc           (1)
  //, var_decay         (1)
  //, order             (assigns, activity)
  , learnDB           (assigns)
  , last_simplify     (-1)
  , progress_estimate (0)
  , verbosity         (0) {
  //conc.clear();
  if (signal(SIGALRM, sig_alrm) == SIG_ERR) {
    fprintf(stderr,"c signal (SIGALRM) error\n");
    exit(1);
  }
}

//=================================================================================================
// Minor methods:


// Creates a new SAT variable in the solver. If 'decision_var' is cleared, variable will not be
// used as a decision variable (NOTE! This has effects on the meaning of a SATISFIABLE result).
//
Var Solver::newVar(bool decision_var)
{
  int     index;
  index = nVars();
  watches .push();          // (list for positive literal)
  watches .push();          // (list for negative literal)
  undos   .push();
  reason  .push(NULL);
  assigns .push(toInt(l_Undef));
  level   .push(-1);
  learnDB .newVar(decision_var);
  // learnDB.activity.push(0);
  // learnDB.order.newVar(decision_var);
  return index;
}


// Returns FALSE if immediate conflict.
bool Solver::assume(Lit p) {
    assert(propQ.size() == 0);
    if (verbosity >= 5) fprintf(stderr,L_IND"assume("L_LIT")\n", L_ind, L_lit(p));
    trail_lim.push(trail.size());
    return enqueue(p); }


// Revert one variable binding on the trail.
//
inline void Solver::undoOne(void)
{
    if (verbosity >= 5){ Lit p = trail.last(); fprintf(stderr,L_IND"unbind("L_LIT")\n", L_ind, L_lit(p)); }
    Lit     p  = trail.last(); trail.pop();
    Var     x  = var(p);
    assigns[x] = toInt(l_Undef);
    reason [x] = NULL;
    level  [x] = -1;
    learnDB.undo(x);
    while (undos[x].size() > 0)
        undos[x].last()->undo(*this, p),
        undos[x].pop();
}


// Reverts to the state before last 'assume()'.
//
void Solver::cancel(void)
{
    assert(propQ.size() == 0);
    if (verbosity >= 5){ if (trail.size() != trail_lim.last()){ Lit p = trail[trail_lim.last()]; fprintf(stderr,L_IND"cancel("L_LIT")\n", L_ind, L_lit(p)); } }
    for (int c = trail.size() - trail_lim.last(); c != 0; c--)
        undoOne();
    trail_lim.pop();
}


// Revert to the state at given level.
//
void Solver::cancelUntil(int level) {
    while (decisionLevel() > level) cancel(); }


// Record a clause and drive backtracking. 'clause[0]' must contain the asserting literal.
//
void Solver::record(const vec<Lit>& clause)
{
    assert(clause.size() != 0);
    LClause* c;
    check(LClause_new(*this, clause, c));
    check(ok);
    check(enqueue(clause[0], c));
    if (c != NULL) learnDB.add(c);
}


//=================================================================================================
// Major methods:


/*_________________________________________________________________________________________________
|                                                                                                  
|  analyze : (confl : BaseConstraint*) (out_learnt : vec<Lit>&) (out_btlevel : int&)  ->  [void]           
|                                                                                                  
|  Description:                                                                                    
|    Analyze conflict and produce a reason clause.                                                 
|                                                                                                  
|    Pre-conditions:                                                                               
|      * 'out_learnt' is assumed to be cleared.                                                    
|      * Current decision level must be greater than root level.                                   
|                                                                                                  
|    Post-conditions:                                                                              
|      * 'out_learnt[0]' is the asserting literal at level 'out_btlevel'.                          
|                                                                                                  
|  Effect:                                                                                         
|    Will undo part of the trail, upto but not beyond the assumption of the current decision level.
|________________________________________________________________________________________________@*/
void Solver::analyze(BaseConstraint* confl, vec<Lit>& out_learnt, int& out_btlevel)
{
  vec<char>&  seen = analyze_seen;
  int         pathC    = 0;
  Lit         p = lit_Undef;
  vec<Lit>    p_reason;
  seen.growTo(nVars(), 0);
  
  // Generate conflict clause:
  //
  out_learnt.push();      // (leave room for the asserting literal)
  out_btlevel = 0;
  do{
    assert(confl != NULL);          // (otherwise should be UIP)
    
    p_reason.clear();
    confl->calcReason(*this, p, p_reason);
    
    if(verbosity >= 5 && p_reason.size() > 0) fprintf(stderr,L_IND"Looking at", L_ind);
    for (int j = 0; j < p_reason.size(); ++j){
      Lit q = p_reason[j];
      assert(var(q) <= nVars());
      assert(var(q) >= 0);
      if (!seen[var(q)]){
	seen[var(q)] = 1;
	if(verbosity >= 5) fprintf(stderr, " "L_LIT"(%d)",L_lit(q),level[var(q)]);
	if (level[var(q)] == decisionLevel())
	  pathC++;
	else{
	  out_learnt.push(~q),
	    out_btlevel = max(out_btlevel, level[var(q)]);
	}
      }
    }
    if(verbosity >= 5) fprintf(stderr,"\n");
    
    // Select next clause to look at:
    do{
      p = trail.last();
      assert(var(p) <= nVars());
      assert(var(p) >= 0);
      confl = reason[var(p)];
      undoOne();
    }while (!seen[var(p)]);
    pathC--;
    assert(pathC >= -1);
    seen[var(p)] = 0;
  }while (pathC > 0);
  
  out_learnt[0] = ~p;
  
  for (int j = 0; j < out_learnt.size(); j++) seen[var(out_learnt[j])] = 0;    // ('seen[]' is now cleared)
  
  if (verbosity >= 5){
    fprintf(stderr,L_IND"Learnt {", L_ind);
    for (int i = 0; i < out_learnt.size(); i++) fprintf(stderr," "L_LIT, L_lit(out_learnt[i]));
    fprintf(stderr," } at level %d\n", out_btlevel); }
  
}


/*_________________________________________________________________________________________________
|                                                                                                  
|  enqueue : (p : Lit) (from : BaseConstraint*)  ->  [bool]                                                
|                                                                                                  
|  Description:                                                                                    
|    Puts a new fact on the propagation queue as well as immediately updating the variable's value.
|    Should a conflict arise, FALSE is returned.                                                   
|                                                                                                  
|  Input:                                                                                          
|    p    - The fact to enqueue                                                                    
|    from - [Optional] Fact propagated from this (currently) unit clause. Stored in 'reason[]'.    
|           Default value is NULL (no reason).                                                     
|                                                                                                  
|  Output:                                                                                         
|    TRUE if fact was enqueued without conflict, FALSE otherwise.                                  
|________________________________________________________________________________________________@*/
bool Solver::enqueue(Lit p, BaseConstraint* from)
{
   if (value(p) != l_Undef){
        if (value(p) == l_False){
            // Conflicting enqueued assignment
            if (decisionLevel() == 0)
                ok = false;
            return false;
        }else{
            // Existing consistent assignment -- don't enqueue
            return true;
        }
      }else{
        // New fact -- store it.
        if (verbosity >= 5) fprintf(stderr,L_IND"bind("L_LIT")\n", L_ind, L_lit(p));
        assigns[var(p)] = toInt(lbool(!sign(p)));
        level  [var(p)] = decisionLevel();
        reason [var(p)] = from;
        trail.push(p);
        propQ.insert(p);
        return true;
    }
}


/*_________________________________________________________________________________________________
|                                                                                                  
|  propagate : [void]  ->  [BaseConstraint*]                                                               
|                                                                                                  
|  Description:                                                                                    
|    Propagates all enqueued facts. If a conflict arises, the conflicting clause is returned,      
|    otherwise NULL.                                                                               
|                                                                                                  
|    Post-conditions:                                                                              
|      * the propagation queue is empty, even if there was a conflict.                             
|________________________________________________________________________________________________@*/
BaseConstraint* Solver::propagate(void)
{
    BaseConstraint* confl = NULL;
    Lit p;

    while (propQ.size() > 0){
        ++stats.propagations;
        p  = propQ.dequeue();        // 'p' is enqueued fact to propagate.
	int ip = index(p);
	vec<BaseConstraint*>& ws = watches[ip];
        bool          keep_watch;
        int           i, j;

        for (i = j = 0; confl == NULL && i < ws.size(); ++i){
            ++stats.inspects;
            keep_watch = false;
            if (!ws[i]->propagate(*this, p, keep_watch))
                confl = ws[i], propQ.clear();
            if (keep_watch) ws[j++] = ws[i];
        }

        // Copy the remaining watches:
        while (i < ws.size())
            ws[j++] = ws[i++];

        watches[ip].shrink(i - j);
    }
    return confl;
}


/*_________________________________________________________________________________________________
|                                                                                                  
|  reduceDB : ()  ->  [void]                                                                       
|                                                                                                  
|  Description:                                                                                    
|    Remove half of the learnt clauses, minus the clauses locked by the current assignment. Locked 
|    clauses are clauses that are reason to a some assignment.                                     
|________________________________________________________________________________________________@*/
void Solver::reduceDB(void)
{
  learnDB.reduceDB(*this);
}


/*_________________________________________________________________________________________________
|                                                                                                  
|  simplifyDB : [void]  ->  [bool]                                                                 
|                                                                                                  
|  Description:                                                                                    
|    Simplify all constraints according to the current top-level assigment (redundant constraints  
|    may be removed altogether).                                                                   
|________________________________________________________________________________________________@*/
void Solver::simplifyDB(void)
{
    if (!ok) return;    // GUARD (public method)
    assert(decisionLevel() == 0);

    if (propagate() != NULL){
        ok = false;
        return; }
    if (nAssigns() == last_simplify)
        return;

    last_simplify = nAssigns();

    learnDB.simplifyDB(*this);
    constDB.simplifyDB(*this);
}

/*_________________________________________________________________________________________________
|                                                                                                  
|  search : (nof_conflicts : int) (nof_learnts : int) (params : const SearchParams&)  ->  [lbool]  
|                                                                                                  
|  Description:                                                                                    
|    Search for a model the specified number of conflicts, keeping the number of learnt clauses    
|    below the provided limit. NOTE! Use negative value for 'nof_conflicts' or 'nof_learnts' to    
|    indicate infinity.                                                                            
|                                                                                                  
|  Output:                                                                                         
|    'l_True' if a partial assigment that is consistent with respect to the clauseset is found. If 
|    all variables are decision variables, this means that the clause set is satisfiable. 'l_False'
|    if the clause set is unsatisfiable. 'l_Undef' if the bound on number of conflicts is reached. 
|________________________________________________________________________________________________@*/
lbool Solver::search(int nof_conflicts, int nof_learnts, const SearchParams& params)
{
    if (!ok) return l_False;    // GUARD (public method)
    assert(root_level == decisionLevel());

    stats.starts++;
    int     conflictC = 0;
    learnDB.setParams(params);

    for (;;){
        BaseConstraint* confl = propagate();
        if (confl != NULL){
            // CONFLICT

            if (verbosity >= 5) fprintf(stderr,L_IND"**CONFLICT in a %s**\n", L_ind, confl->name());
            stats.conflicts++; conflictC++;
            vec<Lit>    learnt_clause;
            int         backtrack_level;
            if (decisionLevel() == root_level)
                return l_False;
            analyze(confl, learnt_clause, backtrack_level);
	    assert(backtrack_level >= 0);
            cancelUntil(max(backtrack_level, root_level));
	    
            record(learnt_clause);
            learnDB.varDecayActivity(); learnDB.claDecayActivity();

        }else{
	  // NO CONFLICT
	  
            if (graceful_quit || (nof_conflicts >= 0 && conflictC >= nof_conflicts)){
	        // graceful_quit or
                // Reached bound on number of conflicts:
                progress_estimate = progressEstimate();
                assert(propQ.size() == 0);
		cancelUntil(root_level);
                return l_Undef; }

            if (decisionLevel() == 0)
                // Simplify the set of problem clauses:
                simplifyDB(), assert(ok);

            if (nof_learnts >= 0 && learnDB.size()-nAssigns() >= nof_learnts)
                // Reduce the set of learnt clauses:
                reduceDB();

            // New variable decision:
            stats.decisions++;
	    Var next = learnDB.select(*this, params.random_var_freq);

            if (next == var_Undef){
                // Model found:
                return l_True;
            }

            check(assume(~Lit(next)));
        }
    }
}


// Return search-space coverage. Not extremely reliable.
//
double Solver::progressEstimate(void)
{
    double  progress = 0;
    double  F = 1.0 / nVars();
    for (int i = 0; i < nVars(); i++)
        if (value(i) != l_Undef)
            progress += pow(F, level[i]);
    return progress / nVars();
}

/*_________________________________________________________________________________________________
|                                                                                                  
|  solve : (assumps : const vec<Lit>&)  ->  [bool]                                                 
|                                                                                                  
|  Description:                                                                                    
|    Top-level solve. If using assumptions (non-empty 'assumps' vector), you must call             
|    'simplifyDB()' first to see that no top-level conflict is present (which would put the solver 
|    in an undefined state).                                                                       
|________________________________________________________________________________________________@*/
lbool Solver::solve(const vec<Lit>& assumps)
{
    simplifyDB();
    if (!ok) return l_False;

    SearchParams    params(0.95, 0.999, 0.02);
    double  nof_conflicts = 100;
    double  nof_learnts   = nConstrs() / 3;
    lbool   status        = l_Undef;

    graceful_quit = 0;
    model.clear();
    model.growTo(nVars());
    for (int i = 0; i < nVars(); i++) model[i] = l_Undef;

    for (int i = 0; i < assumps.size(); i++)
      if (!assume(assumps[i]) || propagate() != NULL){
	propQ.clear();
	cancelUntil(0);
	return l_False; }
    root_level = decisionLevel();

    while (status == l_Undef && graceful_quit == 0){
        if (verbosity >= 1) fprintf(stderr,"c Solving -- conflicts=%d   learnts=%d   progress=%.4f %%\n", (int)nof_conflicts, (int)nof_learnts, progress_estimate*100);
        status = search((int)nof_conflicts, (int)nof_learnts, params);
        nof_conflicts *= 1.5;
        nof_learnts   *= 1.1;
    }
    
    for (int i = 0; i < nVars(); i++) model[i] = value(i);
    cancelUntil(0);
    return status;
}

/*_________________________________________________________________________________________________
|                                                                                                  
|  UPL : (assumps : const vec<Lit>&)  ->  [bool]                                                 
|                                                                                                  
|  Description:                                                                                    
|    Top-level UPL. If using assumptions (non-empty 'assumps' vector), you must call             
|    'simplifyDB()' first to see that no top-level conflict is present (which would put the solver 
|    in an undefined state).                                                                       
|________________________________________________________________________________________________@*/
lbool Solver::UPL(const vec<Lit>& assumps)
{
    simplifyDB();
    if (!ok) return l_False;
    model.clear();

    for (int i = 0; i < assumps.size(); i++)
      if (!assume(assumps[i]) || propagate() != NULL){
	propQ.clear();
	cancelUntil(0);
	return l_False; }

    model.growTo(nVars());
    for (int i = 0; i < nVars(); i++) model[i] = value(i);
    cancelUntil(0);
    return l_True;
}
