/**************************************************************************************************

Solver.h -- (C) Niklas Een, Niklas Srensson, 2004

A simple Chaff-like SAT-solver with support for incremental SAT.

**************************************************************************************************/

#ifndef Solver_h
#define Solver_h

#include <ctime>
#include <signal.h>
#include "SolverTypes.h"
#include "Constraints.h"
#include "Queue.h"
#include "VarOrder.h"
#include "ConstraintsDB.h"

#include "debug.h"

#define NPROP 10

//=================================================================================================
// Solver -- the main class:


struct SolverStats {
  int64   starts, decisions, propagations, inspects, conflicts, propm, propx, propi, propa, propj, prop[NPROP+1];
  SolverStats(void) : starts(0), decisions(0), propagations(0), inspects(0), conflicts(0),
    propm(0), propx(0), propi(0), propa(0), propj(0)
  {
    for(int i = 0; i < NPROP; ++i) prop[i] = 0;
  }
};

class Solver {
protected:
    bool                ok;             // If FALSE, the constraints are already unsatisfiable. No part of the solver state may be used!
    ConstraintsDB<BaseConstraint*> constDB;
    ConstraintsDB<LClause*>        learnDB;

    vec<vec<BaseConstraint*> > watches; // 'watches[lit]' is a list of constraints watching 'lit' (will go there if literal becomes true).
    vec<vec<BaseConstraint*> > undos;   // 'undos[var]' is a list of constraints that will be called when 'var' becomes unbound.
    Queue<Lit>          propQ;          // Propagation queue.

    vec<char>           assigns;        // The current assignments (lbool:s stored as char:s).
    vec<Lit>            trail;          // List of assignments made. 
    vec<int>            trail_lim;      // Separator indices for different decision levels in 'trail'.
    vec<BaseConstraint*> reason;        // 'reason[var]' is the clause that implied the variables current value, or 'NULL' if none.
    vec<int>            level;          // 'level[var]' is the decision level at which assignment was made.
    int                 root_level;     // Level of first proper decision.
    int                 last_simplify;  // Number of top-level assignments at last 'simplifyDB()'.

    // Temporaries (to reduce allocation overhead):
    //
    vec<char>           analyze_seen;

    // Main internal methods:
    //
    bool        assume       (Lit p);
    void        undoOne      (void);
    void        cancel       (void);
    void        cancelUntil  (int level);
    void        record       (const vec<Lit>& clause);

    void        analyze      (BaseConstraint* confl, vec<Lit>& out_learnt, int& out_btlevel); // (bt = backtrack)
    bool        enqueue      (Lit fact, BaseConstraint* from = NULL);
    BaseConstraint* propagate    (void);
    void        reduceDB     (void);
    Lit         pickBranchLit(const SearchParams& params);
    lbool       search       (int nof_conflicts, int nof_learnts, const SearchParams& params);
    double      progressEstimate(void);

    int     decisionLevel(void) { return trail_lim.size(); }

public:

    Solver(void);

    ~Solver(void) {
        for (int i = 0; i < learnDB.size(); i++) learnDB.remove(*this, i, true);
        for (int i = 0; i < constDB.size(); i++) constDB.remove(*this, i, true); }

    int lenWatches(Lit l) const { return watches[index(l)].size(); }

    // Helpers: (semi-internal)
    //
    lbool   value(Var x) { return toLbool(assigns[x]); }
    lbool   value(Lit p) { return sign(p) ? ~toLbool(assigns[var(p)]) : toLbool(assigns[var(p)]); }

    int     nAssigns(void) { return trail.size(); }
    int     nConstrs(void) { return constDB.size(); }
    int     nLearnts(void) { return learnDB.size(); }

    // Statistics: (read-only member variable)
    //
    SolverStats stats;

    // Problem specification:
    //
    Var     newVar (bool decision_var = true);
    int     nVars  (void)  { return assigns.size(); }
    void    addUnit(Lit p) { if (ok) enqueue(p); }

    // -- constraints:
    friend class Clause;
    friend class LClause;
    friend class SClause;
    friend class AtMost;
    friend class Xor;
    friend class Imply;
    friend class jmf;
    friend class Lex;
    friend bool LClause_new(Solver& S, const vec<Lit>& ps, LClause*& out_clause);
    friend bool SClause_new(Solver& S, const vec<Lit>& ps, SClause*& out_clause);
    friend bool AtMost_new(Solver& S, const vec<Lit>& ps, int  max   , AtMost*& out_constr);
    friend bool Xor_new(Solver& S, const vec<Lit>& ps, Xor*& out_constr);
    friend bool Imply_new(Solver& S, const Lit c, const vec<Lit>& ps, Imply*& out_constr);
    friend bool jmf_new(Solver& S, const vec<Lit>& ps, const vec<Lit>& qs, jmf*& out_constr);
    friend bool Lex_new(Solver& S, const vec<Lit>& ps, const vec<Lit>& qs, Lex*& out_constr);

    // void    addClause(const vec<Lit>& ps)        { if (ok){ Clause* c; ok = Clause_new(*this, ps, false, c); if (c != NULL) constDB.add(c); } }
    void    addSClause(const vec<Lit>& ps)       { if (ok){ SClause* c; ok = SClause_new(*this, ps, c); if (c != NULL) constDB.add(c); } }
    void    addAtMost(const vec<Lit>& ps, int n) { if (ok){ AtMost* c; ok = AtMost_new(*this, ps, n    , c); if (c != NULL) constDB.add(c); } }
    void    addXor   (const vec<Lit>& ps)        { if (ok){ Xor*    c; ok = Xor_new   (*this, ps       , c); if (c != NULL) constDB.add(c); } }
    void    addImply (const Lit l, const vec<Lit>& ps) { if (ok){ Imply* c; ok = Imply_new(*this, l, ps, c); if (c != NULL) constDB.add(c); } }
    void    addjmf   (const vec<Lit>& ps, const vec<Lit>& qs) { if (ok){ jmf* c; ok = jmf_new(*this, ps, qs, c); if (c != NULL) constDB.add(c); } }
    void    addLex   (const vec<Lit>& ps, const vec<Lit>& qs) { if (ok){ Lex* c; ok = Lex_new(*this, ps, qs, c); if (c != NULL) constDB.add(c); } }

    // Solving:
    //
    bool    okay(void) { return ok; }
    void    simplifyDB(void);
    lbool   solve(const vec<Lit>& assumps = vec<Lit>());
    lbool   UPL(const vec<Lit>& assumps = vec<Lit>());

    vec<int> getOrder() { return learnDB.getOrder(*this); }

    double      progress_estimate;  // Set by 'search()'.
    vec<lbool>  model;              // If problem is solved, this vector contains the model (if any).
    int         verbosity;          // Verbosity level. 0=silent, 1=some progress report, 2=everything

    //vec<vec<Lit>* > conc;              // Contains short clauses learnt in the solvingprocess.
};

//=================================================================================================

#endif
